#include "Chromosome.h"

//using namespace std;



// ================================================================
// GenomeMixer    (c) 2003, Alex Williams
// All rights reserved.
// Send email correspondence to Robert W. Williams <rwilliam@nb.utmem.edu>
// at the University of Tennesse Health Science Center

// This project was funded jointly by the National Institute of Mental Health, National Institute on Drug Abuse,
// and the National Science Foundation (award P20-MH 62009 to KFM and RW)

// This source code may be used according to the terms of the GNU General Public License (GPL).
// See www.gnu.org for specifics. Essentially, the source may be freely used as long as the source code
// for whatever program the source is also made freely available. If you want to use this source
// in a closed-source program, you should email us for licensing options.

// Disclaimer: This code comes with no warranty at all, and there is no guarantee that it will
// work as claimed in the documentation. Hopefully, it will be useful anyway.

// Note that GenomeMixer makes use of the Trolltech Qt toolkit for its window-managing needs.
// There is very little use of Qt data structures (aside from QString) in the actual processing code,
// but **you will need to have Qt installed in order to compile this code.** Check <www.trolltech.com>
// for details; there should be a free GPL-software-only version of Qt available on the Trolltech
// web site.

// Any redistributions of this software in code or binary form must include the
// above copyright notice, licensing information, disclaimer, and this notice itself.
// Basically, all the comments at the top of this file must be retained.
// ================================================================


int Chromosome::myNumChrWithXY = 0;
int Chromosome::myXChrIndex = -1; // the index of the X chromosome (since it is handled differently)
int Chromosome::myYChrIndex = -1; // the index of the Y chromosome (since it is handled differently)
std::vector <string>Chromosome::chrNames;//(10);
std::vector <float>Chromosome::chrMBaseLength;//(10);
std::vector <float>Chromosome::chr_cM_Length;//(10);


Chromosome::Chromosome() {	// Constructor
    thisChrNum = -1;
    genes.clear();
    genOfRecombination = 0;
}

Chromosome::Chromosome(int theNum) {	// Constructor
    thisChrNum = theNum;
    genes.clear();
    genOfRecombination = 0;
}

void Chromosome::addCopyOfGR(GeneRange *newGeneRange) {
    // Copies the relevant data from this GR and adds it
    addGR(newGeneRange->start(), newGeneRange->end(), newGeneRange->value());
}

void Chromosome::addGR(long start, long end, int setTo) {
    if (genes.size() == genes.capacity()) {
        genes.reserve(genes.capacity()*2 + 5); // resize so that we have enough capacity...
    }
    genes.push_back(new GeneRange(start, end, setTo));
}

void Chromosome::addPurebredGR(int setTo) {
    addGR(0, lengthBases(), setTo); // just makes a whole chromosome of "setTo"
}

GeneRange* Chromosome::getGRfromActual(long findActual) {
    long bottom = 0;
    long top    = genes.size() - 1;
    long middle = int( top / 2);
        
    while (1) {		// infinite loop (binary search)--we exit with a "return" statement
        GeneRange *getItem = getGRfromIndex(middle); // we'll see if this is what we're loking for
        
        if (findActual < getItem->start()) {
            // we're looking for something that's before the start of the middle item
            
            if (middle == 0) {
                cout << "Can't find gene range with index " << findActual << "\n. Exiting...";
                exit(1);
                //return NULL; // <-- we failed to find the GR; it doesn't exist in this chromosome
            }
            top = (middle - 1); // the old middle is now the top of the range
            middle = int( (top + bottom)/2 );
            // now the middle is halfway in between the 
            // previous middle and the bottom (round down)
                
        } else if (findActual > getItem->end()) {
            // we're looking for something that's after the end of the middle item
            
            if (middle == long(genes.size() - 1) ) {
                cout << "Can't find gene range with index " << findActual << "\n. Exiting...";
                exit(1);
                //return NULL; // <-- we failed to find the GR; it doesn't exist in this chromosome
            }
            bottom = (middle + 1);
            middle = int( (top + bottom)/2 );
            // now the middle is halfway between the 
            // previous middle and the bottom (round down)
            
        } else {
            return getItem;
            // well, it's not after this item, and it's not before
            // this item, so we must be looking for this item
        }
    }
}



GeneRange* Chromosome::getGRfromIndex(long thisIndex) {
    // Get a GeneRange from its index in the vector.
    // If we can't find a suitable GR, it exits the program with an error message. 
    
    if (thisIndex < 0 || thisIndex >= long(genes.size()) ) {
	cerr << endl << endl << 
		"ERROR in Chromosome::getGRfromIndex. Invalid value for thisIndex: thisIndex is " << thisIndex << ", genes.size() is " << genes.size() << "." << endl << endl;
	assert( (thisIndex >= 0) && (thisIndex < long(genes.size())) );
	exit(1);
    }
    
    return genes[thisIndex];
    //return genes.at(thisIndex); <-- ".at(...)" generates an error for some reason (but only on the Linux/QT machine): no matching function for call to `vector<GeneRange *, allocator<GeneRange *>::at (long int &)'
}



long Chromosome::getIndexFromActual(long thisGeneNum) {
    for (unsigned long i = 0; i < genes.size(); i++) {
        // this should be a binary search
        GeneRange *gr = getGRfromIndex(i);
        if (thisGeneNum >= gr->start() && thisGeneNum <= gr->end()) {
            return i;
        }
    }
    
    cerr << "ERROR in a call to Chromosome::getIndexFromActual. This function failed to find any GeneRange that contained the actual (base) value \"" << thisGeneNum << "\". genes.size() is " << genes.size() << "." << endl;
    printGeneRanges();
    
    assert("Error" == "Chromosome::getIndexFromActual");
    exit(1);
}


void Chromosome::printGeneRanges() {
    cout << "Here are the gene ranges for chromosome \"" << name() << "\" (length: " << lengthBases() << " bases, " << length_cM() << " cM):" << endl;
    
    for (unsigned long i = 0; i < genes.size(); i++) {
	GeneRange *gr = getGRfromIndex(i);
	cout << "Value: " << gr->value() << ", start: " << gr->start() << ", end: " << gr->end() << endl;
    }

}

void Chromosome::setNum(int newNum) { 
  thisChrNum = newNum; 
}


bool Chromosome::isFemSexChr() {
    return (Chromosome::xChrIndex() == thisChrNum);
    // returns true if this is the X chromosome, false otherwise
}

bool Chromosome::isMaleSexChr() {
    return (Chromosome::yChrIndex() == thisChrNum);
    // returns true if this is the Y chromosome, false otherwise
}


int     Chromosome::numberOfChr() {	
  return chrNames.size();
}

string  Chromosome::name() {	
  return chrNameFromIndex(thisChrNum);
}

int     Chromosome::num() {		
  return thisChrNum;		
}

float   Chromosome::lengthMBASE() {	
  return lengthMBASEfromIndex(thisChrNum);
}

long    Chromosome::lengthBases() {	
  return lengthBasesFromIndex(thisChrNum);
}

float   Chromosome::length_cM()   {        
 return length_cM_fromIndex(thisChrNum);
}

unsigned long    Chromosome::numGeneRanges() {
  return genes.size();		 
}





// ========= Class-wide methods below =========
void Chromosome::printClassData(void) {
    cout << "Chromosome class-wide data:" << endl;
    for (int i = 0; i < Chromosome::numChr(); i++) {
        cout << " Chromosome index " << i << ", name \"" << (Chromosome::chrNameFromIndex(i)) << "\", MBase " << (Chromosome::lengthMBASEfromIndex(i)) << ", and cM length " << (Chromosome::length_cM_fromIndex(i)) << endl;
    }
}



void Chromosome::addChrData(string chrName, float chr_MB, float chr_cM) {
    (Chromosome::chrNames.push_back(string(chrName)));
    (Chromosome::chrMBaseLength.push_back(chr_MB));
    (Chromosome::chr_cM_Length.push_back(chr_cM));
    assert(long(Chromosome::chrNames.size()) <= Chromosome::myNumChrWithXY);
}

int    Chromosome::numChr() {
  assert(myNumChrWithXY >= 0);
  
  return myNumChrWithXY;
}

int Chromosome::xChrIndex() {
  assert(myXChrIndex >= 0 && myXChrIndex < myNumChrWithXY);
  return myXChrIndex;
}

int Chromosome::yChrIndex() {
  assert(myYChrIndex >= 0 && myYChrIndex < myNumChrWithXY);
  return myYChrIndex; 
}


float Chromosome::allChrTotal_cM(void) {
  float total = 0.0;
  for (int chr = 0; chr < Chromosome::numChr(); chr++) {
    total += Chromosome::length_cM_fromIndex(chr);
  }
  return total;
}


long Chromosome::allChrTotalBases(void) {
  long total = 0;
  for (int chr = 0; chr < Chromosome::numChr(); chr++) {
    total += Chromosome::lengthBasesFromIndex(chr);
  }
  return total;
}


void Chromosome::setNumberOfChromosomes(int numChr) {
    (Chromosome::myNumChrWithXY) = numChr;
    (Chromosome::myXChrIndex) = numChr - 2;
    (Chromosome::myYChrIndex) = numChr - 1;    
    // Below: sets the vectors to the size that we are going to need,
    // so they don't keep resizing over and over (this is just for efficiency.
    // Things would still work if these three lines were removed).
    (Chromosome::chrNames).reserve(numChr);
    (Chromosome::chrMBaseLength).reserve(numChr);
    (Chromosome::chr_cM_Length).reserve(numChr);
    
    (Chromosome::chrNames).clear();
    (Chromosome::chrMBaseLength).clear();
    (Chromosome::chr_cM_Length).clear();
}

void Chromosome::clearAllChrData(void) {
  Chromosome::setNumberOfChromosomes(0);
  (Chromosome::chrNames).clear();
  (Chromosome::chrMBaseLength).clear();
  (Chromosome::chr_cM_Length).clear();
}


string Chromosome::chrNameFromIndex(int chrNum) { // <-- Get the name of the chr at index "chrNum"
    assert(chrNum >= 0 && chrNum < Chromosome::numChr()); // <-- respect array bounds
    return Chromosome::chrNames[chrNum];
}

float Chromosome::lengthMBASEfromIndex(int chrNum) {
    assert(chrNum >= 0 && chrNum < Chromosome::numChr()); // <-- respect array bounds
    return Chromosome::chrMBaseLength[chrNum];
}

long Chromosome::lengthBasesFromIndex(int chrNum) {
    // Note that this ends up as just a call to the class method "lengthMBASEfromIndex"
    return long(1000000 * lengthMBASEfromIndex(chrNum)); // 1 million times the megabase value
}

float Chromosome::length_cM_fromIndex(int chrNum) {
    assert(chrNum >= 0 && chrNum < Chromosome::numChr()); // <-- respect array bounds
    return Chromosome::chr_cM_Length[chrNum];
}
// ========= Class-wide methods above ========= 




